# coding=utf-8
"""
Manages YAML configuration
"""

import collections
import os

import everett
import everett.manager
import yaml


def update_nested_dict(target_dict: collections.MutableMapping,
                       source_dict: collections.MutableMapping) -> collections.MutableMapping:
    """
    Updates a config dictionary that might contain other config dictionaries
    :param target_dict: dictionary to update
    :param source_dict: dictionary that contains values to use
    :return: updated target_dict
    """
    if source_dict is None:
        return target_dict
    for key, value in source_dict.items():
        if isinstance(value, collections.MutableMapping):
            target_dict[key.upper()] = update_nested_dict(
                target_dict.get(key, {}), value)
        else:
            target_dict[key.upper()] = source_dict[key]
    return target_dict


def flatten_dict(source_dict: collections.MutableMapping, parent_key='', sep='_') -> dict:
    """
    Flattens a dictionary, adding namespaces to key names
    :param source_dict: dictionary to flatten
    :param parent_key: namespace
    :param sep: separator to use, defaults to "_"
    :return: config dictionary
    """
    items = []
    for key, value in source_dict.items():
        new_key = parent_key + sep + key if parent_key else key
        if isinstance(value, collections.MutableMapping):
            items.extend(flatten_dict(value, new_key, sep=sep).items())
        else:
            items.append((new_key, value))
    return dict(items)


class YAMLConfig:
    """
    Represents a Config coming from a YAML file
    """

    def __init__(self, possible_paths):
        self.cfg = {}
        possible_paths = everett.manager.listify(possible_paths)

        for path in possible_paths:
            if not path:
                continue

            path = os.path.abspath(os.path.expanduser(path.strip()))
            if path and os.path.isfile(path):
                self.cfg = update_nested_dict(
                    self.cfg, self.parse_yaml_file(path))

        self.cfg = flatten_dict(self.cfg)

    @staticmethod
    def parse_yaml_file(path: str) -> dict:
        """
        Reads a YAML file into a dictionary
        :param path: path to YAML file
        :return: dictionary
        """
        with open(path) as stream:
            return yaml.load(stream)

    def get(self, key, namespace=None):
        """
        Obtain a key value from the config dict
        :param key: key name
        :param namespace: namespace of the config key
        :return: config value
        """
        value = everett.manager.get_key_from_envs(self.cfg, key, namespace)
        return value



# coding=utf-8
"""
Tests random package
"""

from hypothesis import given
from hypothesis import strategies as st

from elib.custom_random import random_bytes, random_string


def test_random_str():
    res = set()
    for _ in range(10000):
        rand = random_string()
        assert len(rand) == 6
        assert rand not in res
        res.add(rand)


def test_random_bytes():
    res = set()
    for _ in range(100000):
        rand = random_bytes()
        assert len(rand) == 1024
        assert rand not in res
        res.add(rand)


@given(length=st.integers(min_value=1, max_value=1024))
def test_random_string_length(length):
    assert len(random_string(length)) == length



# coding=utf-8
"""
Tests paste package
"""

import pytest
from httmock import HTTMock, all_requests, response

from elib.paste import PasteContent, PasteError, create_new_paste


@all_requests
def correct_response(*_):
    """Returns a correct response"""
    return response(200, content={'url': 'content'})


@all_requests
def bad_response(*_):
    """Returns  a bad response"""
    return response(500, content={'reason': 'reason'})


def test_pastebin_correct():
    with HTTMock(correct_response):
        content = PasteContent('test.lua', """print('caribou')""")
        resp = create_new_paste('test', [content])
        assert resp == 'content'


def test_pastebin_bad():
    with HTTMock(bad_response):
        content = PasteContent('test.lua', """print('caribou')""")
        with pytest.raises(PasteError):
            create_new_paste('test', [content])



# coding=utf-8
"""
Tests TTS package
"""

from pathlib import Path

import pytest

from elib.tts import text_to_speech

# from httmock import all_requests, response
# from requests import PreparedRequest


#
# @all_requests
# def correct_response(url, request: PreparedRequest):
#     if request.url == 'https://translate.google.com/':
#         return response(
#             200,
#             content={'text': r";MAX_ALTERNATIVES_ROUNDTRIP_RESULTS=1;"
#                              r"TKK=eval('((function(){var a\x3d1869649412;var b\x3d1066560713;"
#                              r"return 420586+\x27.\x27+(a+b)})())');WEB_TRANSLATION_PATH="})
#     print(url, request)
#     return response(200, content={'url': 'content'})
#
#
# @all_requests
# def bad_response(*_):
#     return response(500, content={'reason': 'reason'})


def test_tts_correct():
    file = Path('test.file')
    assert not file.exists()
    text_to_speech('some text', file)
    assert (file.exists())
    with pytest.raises(FileExistsError):
        text_to_speech('some text', file)



# coding=utf-8

import string

import delegator
import pexpect
import pytest
from hypothesis import given
from hypothesis import strategies as st
from mockito import mock, unstub, verify, when

import elib._run
from elib._run import filter_line


@pytest.fixture(name='process', autouse=True)
def _process():
    process = mock()
    subprocess = mock()
    process.out = 'output'
    process.err = ''
    process.name = 'test.exe'
    process.return_code = 0
    process.subprocess = subprocess
    when(delegator).run(...).thenReturn(process)
    when(elib._run).find_executable('test').thenReturn(process)
    when(elib._run).cmd_start(...)
    when(elib._run).cmd_end(...)
    when(elib._run).info(...)
    when(elib._run).error(...)
    when(elib._run).std_out(...)
    when(elib._run).std_err(...)
    yield process


@given(text=st.text(alphabet=string.printable))
def test_filter_line_raw(text):
    assert filter_line(text, None) == text


def test_filter_line():
    text = 'some random text'
    assert filter_line(text, None) == text
    assert filter_line(text, ['some']) is None
    assert filter_line(text, [' some']) == text
    assert filter_line(text, ['some ']) is None
    assert filter_line(text, ['random']) is None
    assert filter_line(text, [' random']) is None
    assert filter_line(text, ['random ']) is None
    assert filter_line(text, [' random ']) is None
    assert filter_line(text, ['text']) is None
    assert filter_line(text, [' text']) is None
    assert filter_line(text, [' text ']) == text


def test_exe_not_found():
    when(elib._run).find_executable(...).thenReturn(None)
    with pytest.raises(SystemExit):
        elib._run.run('test')


def _basic_check(output, code):
    verify(elib._run).info('test.exe -> 0')
    verify(elib._run, times=0).cmd_start(...)
    verify(elib._run, times=0).cmd_end(...)
    verify(elib._run, times=0).std_err(...)
    verify(elib._run).std_out(output)
    verify(elib._run, times=0).error(...)
    assert code == 0


@pytest.mark.parametrize(
    'input_,output',
    [
        ('test', 'test'),
        ('test\n', 'test'),
        ('test\n\n', 'test'),
        ('test\n\ntest', 'test\ntest'),
        ('test\n\ntest\n', 'test\ntest'),
        ('test\n\ntest\n\n', 'test\ntest'),
    ]
)
def test_output(process, input_, output):
    process.out = input_
    out, code = elib._run.run('test')
    _basic_check(out, code)
    assert out == output


def test_no_output(process):
    process.out = ''
    out, code = elib._run.run('test')
    _basic_check(out, code)
    assert out == ''


def test_filtered_output():
    out, code = elib._run.run('test', filters=['output'])
    _basic_check(out, code)
    assert out == ''


def test_mute_output():
    out, code = elib._run.run('test', mute=True)
    verify(elib._run, times=0).info(...)
    verify(elib._run, times=0).error(...)
    verify(elib._run).cmd_end(' -> 0')
    verify(elib._run, times=0).std_err(...)
    verify(elib._run, times=0).std_out(...)
    assert code == 0
    assert out == 'output'


def test_filter_as_str(process):
    process.out = 'output\ntest'
    out, code = elib._run.run('test', mute=True, filters='test')
    verify(elib._run, times=0).info(...)
    verify(elib._run, times=0).error(...)
    verify(elib._run).cmd_end(' -> 0')
    verify(elib._run, times=0).std_err(...)
    verify(elib._run, times=0).std_out(...)
    assert code == 0
    assert out == 'output'


def test_error(process):
    process.return_code = 1
    process.out = 'some error'
    out, code = elib._run.run('test', filters=['output'], failure_ok=True)
    verify(elib._run, times=0).cmd_start(...)
    verify(elib._run, times=0).cmd_end(...)
    verify(elib._run).std_err('test.exe error:\nsome error')
    verify(elib._run, times=0).std_out(...)
    verify(elib._run, times=1).error('command failed: test.exe -> 1')
    assert code == 1
    assert out == 'some error'


def test_error_no_result(process):
    process.return_code = 1
    process.out = ''
    out, code = elib._run.run('test', filters=['output'], failure_ok=True)
    verify(elib._run, times=0).cmd_start(...)
    verify(elib._run, times=0).cmd_end(...)
    verify(elib._run, times=0).std_err(...)
    verify(elib._run, times=0).std_out(...)
    verify(elib._run).error('command failed: test.exe -> 1')
    assert code == 1
    assert out == ''


def test_error_muted(process):
    process.return_code = 1
    unstub(process.subprocess)
    when(process.subprocess).read_nonblocking(1, None).thenReturn('some output').thenRaise(pexpect.exceptions.EOF(None))
    out, code = elib._run.run('test', filters=['output'], failure_ok=True, mute=True)
    verify(elib._run, times=0).info(...)
    verify(elib._run).cmd_end('')
    verify(elib._run, times=0).std_err(...)
    verify(elib._run, times=0).std_out(...)
    verify(elib._run).error('command failed: test.exe -> 1')
    assert code == 1
    assert out == ''


def test_failure(process):
    process.return_code = 1
    process.out = 'error'
    with pytest.raises(SystemExit):
        elib._run.run('test', filters=['output'])
    verify(elib._run, times=0).cmd_start(...)
    verify(elib._run, times=0).cmd_end(...)
    verify(elib._run).std_err('test.exe error:\nerror')
    verify(elib._run, times=0).std_out(...)
    verify(elib._run).error('command failed: test.exe -> 1')



# coding=utf-8
"""
Install a handler to redirect all INFO messages (and higher) to the Discord Channel
"""
import abc
import logging as base

from elib.custom_logging._constants import LOGGERS


class CustomLoggingHandler(base.Handler):
    """
    Install a handler to redirect all INFO messages (and higher) to the Discord Channel
    """

    def __init__(self, name: str, level=base.INFO):
        """
        Creates a logging Handler
        :param name: name of this handler
        :param level: instance of logging.* level
        """
        base.Handler.__init__(self, level)
        self.set_name(name)

    @abc.abstractmethod
    def emit(self, record: base.LogRecord):
        """
        Redirects the record to the Discord channel if its level is INFO or higher

        Args:
            record: logging.record to emit
        """

    def register(self, logger: base.Logger):
        """
        Registers the handler with an existing logger

        :param logger: logger to attach to
        """
        logger.debug(f'registering logging handler: {self.name}')
        LOGGERS[logger.name][self.name] = self
        logger.addHandler(self)



"""
Placeholder for future personal library
"""
import logging

from pkg_resources import DistributionNotFound, get_distribution

LOGGER = logging.getLogger('ELIB')
LOGGER.setLevel(logging.DEBUG)
LOGGER.addHandler(logging.NullHandler())

# pylint: disable=wrong-import-position
from . import config, console, custom_logging, custom_random, downloader, hash_, paste, path, settings, tts
# pylint: disable=wrong-import-position
from ._run import run
# pylint: disable=wrong-import-position
from .pretty import pretty_format

try:
    __version__ = get_distribution('elib').version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    __version__ = 'not installed'



# coding=utf-8
"""
Convenience functions to manage logging
"""

import logging as base
import logging.handlers as base_handlers
import sys
from pathlib import Path

from elib import LOGGER as ELIB_LOGGER

from . import _constants
from .click_handler import ClickHandler

# noinspection SpellCheckingInspection


def set_handler_level(logger_name, handler_name, level):
    """
    Sets output level for console handlers

    Args:
        logger_name: name of the logger to update
        level: desired level
    """
    if isinstance(level, str):
        level = _constants.LEVELS[level.upper()]
    _constants.LOGGERS[logger_name][handler_name].setLevel(level)


# pylint: disable=too-many-arguments
def _setup_file_logging(logger: base.Logger,
                        logger_name: str,
                        rotate_logs: bool,
                        rotate_log_when: str,
                        rotate_log_backup_count: int,
                        file_level,
                        file_format,
                        **_):
    file_name = Path(f'{logger_name}.log').absolute()
    if rotate_logs:
        file_handler = base_handlers.TimedRotatingFileHandler(
            filename=file_name,
            when=rotate_log_when,
            backupCount=rotate_log_backup_count,
            encoding='utf8',
        )
    else:
        if Path(file_name).exists():
            Path(file_name).unlink()
        file_handler = base.FileHandler(
            filename=file_name,
            encoding='utf8'
        )

    file_handler.setLevel(file_level)
    file_handler.setFormatter(base.Formatter(file_format))

    logger.addHandler(file_handler)
    if rotate_logs:
        logger.debug(f'added rotating file logging handler: {file_name}')
    else:
        logger.debug(f'added file logging handler: {file_name}')

    return file_handler


# pylint: disable= unused-argument
def get_logger(
        logger_name: str,
        log_to_file: bool = False,
        use_click_handler: bool = False,
        rotate_logs: bool = False,
        rotate_log_when: str = 'midnight',
        rotate_log_backup_count: int = 7,
        console_level=base.ERROR,
        file_level=base.DEBUG,
        console_format=_constants.DEFAULT_CONSOLE_FORMAT,
        file_format=_constants.DEFAULT_FILE_FORMAT,
) -> base.Logger:
    """
    Set up the logger

    :param logger_name: name of the logger
    :param log_to_file: path to log file [optional]
    :param use_click_handler: use click handler for console output
    :param rotate_logs: whether or not to rotate the log
    :param rotate_log_when: when log rotation should occur
    :param rotate_log_backup_count: number of log files to keep
    :param console_level: level for console logging
    :param file_level: level for file logging
    :param console_format: console formatter string
    :param file_format: file formatter string

    :return: logger object
    """
    if logger_name in _constants.LOGGERS:
        _constants.LOGGERS[logger_name]['logger'].debug(f'logger already initialized: {logger_name}')
        return _constants.LOGGERS[logger_name]['logger']

    kwargs = locals()

    if _constants.ROOT_LOGGER is None:
        logger = base.getLogger(logger_name)
        _constants.ROOT_LOGGER = logger
    else:
        logger = _constants.ROOT_LOGGER.getChild(logger_name)

    logger.setLevel(base.DEBUG)

    if use_click_handler:
        console_handler = ClickHandler()
    else:
        console_handler = base.StreamHandler(sys.stdout)
    console_handler.setLevel(console_level)
    console_handler.setFormatter(base.Formatter(console_format))

    logger.addHandler(console_handler)
    logger.debug('added console logging handler')

    _constants.LOGGERS[logger.name] = {
        'logger': logger,
        'ch': console_handler,
    }

    if log_to_file:
        file_handler = _setup_file_logging(logger, **kwargs)
        _constants.LOGGERS[logger.name]['fh'] = file_handler

    return logger


def activate_elib_logging():
    """
    Attaches all handlers of the root logger to the ELIB logger
    """
    for handler in _constants.ROOT_LOGGER.handlers:
        ELIB_LOGGER.addHandler(handler)


def get_elib_logger() -> base.Logger:
    """
    Returns ELIB's logger
    """
    return ELIB_LOGGER



# coding=utf-8
"""
Utilities to control the console window
"""

import ctypes

import click

from elib import LOGGER
from elib.settings import ELIBSettings


def set_title(title: str):  # pragma: no cover
    """
    Sets the console title

    Args:
        title: title as a string
    """
    LOGGER.debug(f'setting console title to {title}')
    ctypes.windll.kernel32.SetConsoleTitleW(title)


def _sanitize(input_: str, prefix=True) -> str:
    if prefix:
        input_ = f'EPAB: {input_}'
    return input_.encode('ascii', 'ignore').decode()


def _output(txt, color, **kwargs) -> str:
    if not ELIBSettings.quiet:
        click.secho(txt, fg=color, **kwargs)
    return txt


def info(txt: str, **kwargs):
    """
    Prints out informative text

    :param txt: text to print
    :param kwargs: additional arguments to `click.secho` command
    """
    txt = _sanitize(txt)
    _output(txt, ELIBSettings.color_info, **kwargs)


def error(txt: str, **kwargs):
    """
    Prints out an error

    :param txt: text to print
    :param kwargs: additional arguments to `click.secho` command
    """
    txt = _sanitize(txt)
    _output(txt, ELIBSettings.color_error, err=True, **kwargs)


def cmd_start(txt: str, **kwargs):
    """
    Prints out the start of a sub-process

    This command should be followed by à call to `cmd_end`

    :param txt: text to print
    :param kwargs: additional arguments to `click.secho` command
    """
    txt = _sanitize(txt)
    _output(txt, ELIBSettings.color_cmd, nl=False, **kwargs)


def cmd_end(txt: str, **kwargs):
    """
    Prints out the end of a sub-process

    This should follow a call to `cmd_start`

    :param txt: text to print
    :param kwargs: additional arguments to `click.secho` command
    """
    txt = _sanitize(txt, prefix=False)
    _output(txt, ELIBSettings.color_cmd, **kwargs)


def std_out(txt: str, **kwargs):
    """
    Prints out text from a sub-process standard out stream

    :param txt: text to print
    :param kwargs: additional arguments to `click.secho` command
    """
    txt = _sanitize(f'{txt}', prefix=False)
    _output(txt, ELIBSettings.color_stdout, nl=False, **kwargs)


def std_err(txt: str, **kwargs):
    """
    Prints out text from a sub-process standard error stream

    :param txt: text to print
    :param kwargs: additional arguments to `click.secho` command
    """
    txt = _sanitize(txt, prefix=False)
    _output(txt, ELIBSettings.color_stderr, err=True, **kwargs)



# coding=utf-8
"""
Computes hashes
"""

import hashlib

from elib import LOGGER


def get_hash(data, method: str = 'md5'):
    """
    Computes hash from data

    Args:
        data: data
        method: hash method (defaults to MD5) One of [sha1, sha224, sha256, sha384, sha512, blake2b, blake2s]

    Returns: hash value

    """
    if not isinstance(data, bytes):
        data = bytes(data, 'utf-8')

    try:
        func = getattr(hashlib, method)
    except AttributeError:
        raise RuntimeError('cannot find method "{}" in hashlib'.format(method))
    else:
        hash_ = func(data).hexdigest()
        LOGGER.debug('hash for binary data: %s', hash_)

        return hash_



# coding=utf-8
"""
Logging constants
"""
import logging as base

DEFAULT_CONSOLE_FORMAT = '%(relativeCreated)08d ms ' \
                         '%(levelname)8s ' \
                         '%(name)s ' \
                         '%(message)s'
DEFAULT_FILE_FORMAT = '%(asctime)s %(levelname)8s %(name)s ' \
                      '%(process)d %(processName)s ' \
                      '%(thread)d %(threadName)s ' \
                      '%(pathname)s[%(lineno)d].%(funcName)s: ' \
                      '%(message)s'
LOGGERS = {}
ROOT_LOGGER = None
LEVELS = {
    'DEBUG': base.DEBUG,
    'INFO': base.INFO,
    'WARNING': base.WARN,
    'WARN': base.WARN,
    'ERROR': base.ERROR,
    'ERR': base.ERROR,
    'CRITICAL': base.CRITICAL,
    'CRIT': base.CRITICAL,
}



# coding=utf-8
"""
Path utilities
"""

import typing
from pathlib import Path

from elib import LOGGER


def ensure_path(*path: typing.Union[str, Path], must_exist: bool = True) -> Path:
    """
    Ensure path is a Path instance

    Args:
        path: path
        must_exist: if True, raises FileNotFoundError when path does not exist

    Returns: Path instance
    """
    path = Path(*path).absolute()
    if must_exist and not path.exists():
        raise FileNotFoundError(str(path))
    return path.absolute()


def ensure_file(*file_path: typing.Union[str, Path], must_exist: bool = True) -> Path:
    """
    Ensure path is a Path instance

    Args:
        file_path: path
        must_exist: if True, raises FileNotFoundError when path does not exist

    Returns: Path instance

    """
    file_path = ensure_path(*file_path, must_exist=must_exist)
    if file_path.exists():
        if not file_path.is_file():
            raise TypeError(f'not a file: {str(file_path.absolute())}')
    return file_path


def ensure_dir(*dir_path: typing.Union[str, Path], must_exist: bool = True, create=False) -> Path:
    """
    Ensure path is a Path instance

    Args:
        dir_path: path
        must_exist: if True, raises FileNotFoundError when path does not exist
        create: create the directory if it doesn't exist (implies "must_exist == False")

    Returns: Path instance

    """
    must_exist = not create if create else must_exist
    dir_path = ensure_path(*dir_path, must_exist=must_exist)
    if dir_path.exists():
        if not dir_path.is_dir():
            raise TypeError(f'not a directory: {str(dir_path.absolute())}')
    else:
        if create:
            LOGGER.debug(f'creating directory: {dir_path}')
            dir_path.mkdir(parents=True)
    return dir_path



# coding=utf-8
"""
Formats an object for representation on the console or in a log
"""


import pprint


def _indent(text, indent=4):
    fstring = ' ' * indent + '{}'
    return ''.join([fstring.format(l) for l in text.splitlines(True)])


def pretty_format(obj, indent=4):
    """
    Formats an object for representation on the console or in a log

    Args:
        obj: object to represent
        indent: indent (defaults to 4)

    Returns: formatted text

    """
    return _indent(pprint.pformat(obj, width=120), indent=indent)



# coding=utf-8
"""
Tests logging package
"""

import logging
import logging.handlers
from pathlib import Path

import pytest

import elib
import elib.custom_logging._constants
import elib.custom_logging._custom_logging_handler


@pytest.fixture(scope='function', name='setup_logging')
def _setup_logging(caplog):
    """Provides a logger and the caplog fixture"""
    logger = elib.custom_logging.get_logger(logger_name='TEST_LOGGER', console_level=logging.DEBUG)
    yield logger, caplog


# noinspection PyShadowingNames
def test_logger(setup_logging):
    logger, caplog = setup_logging
    assert isinstance(logger, logging.Logger)
    assert logger is logging.getLogger(logger.name)
    caplog.set_level(logging.DEBUG)
    logger.info('test_message_debug')
    assert 'test_message_debug' in caplog.text
    assert elib.custom_logging.get_logger(logger.name) is logging.getLogger(logger.name)


# noinspection PyShadowingNames
def test_double_instantiation():
    first_logger = elib.custom_logging.get_logger('test_logger')
    second_logger = elib.custom_logging.get_logger('test_logger')
    assert first_logger is second_logger


# noinspection PyShadowingNames
@pytest.mark.parametrize('levels', [
    (logging.DEBUG, logging.INFO, 'debug'),
    (logging.INFO, logging.WARNING, 'info'),
    (logging.WARNING, logging.ERROR, 'warning'),
    (logging.ERROR, logging.CRITICAL, 'error'),
])
def test_logger_levels(setup_logging, levels):
    logger, caplog = setup_logging
    logger.setLevel(levels[1])
    # caplog.set_level(levels[1], logger='TEST_LOGGER')
    out_func = getattr(logger, levels[2])
    out_func('test_message_debug')
    assert 'test_message_debug' not in caplog.text
    logger.setLevel(levels[0])
    # caplog.set_level(levels[0], logger='TEST_LOGGER')
    out_func('test_message_debug')
    assert 'test_message_debug' in caplog.text


# noinspection PyShadowingNames
@pytest.mark.parametrize('level', ['debug', 'info', 'warning', 'error', 'critical'])
def test_logger_not_set(setup_logging, level):
    logger, caplog = setup_logging
    logger.setLevel(1)
    caplog.set_level(logging.NOTSET)
    out_func = getattr(logger, level)
    out_func('test_message_debug')
    assert 'test_message_debug' in caplog.text


def test_file_handler():
    logger_name = elib.custom_random.random_string()
    log_file = Path(f'./{logger_name}.log')
    assert not log_file.exists()
    logger = elib.custom_logging.get_logger(logger_name, log_to_file=True)
    logger.warning('test')
    assert log_file.exists()
    assert 'test' in log_file.read_text()


def test_file_handler_levels():
    logger_name = elib.custom_random.random_string()
    log_file = Path(f'./{logger_name}.log')
    assert not log_file.exists()
    logger = elib.custom_logging.get_logger(logger_name, file_level=logging.ERROR, log_to_file=True)
    logger.warning('__warning__')
    logger.error('__error__')
    assert log_file.exists()
    assert '__warning__' not in log_file.read_text()
    assert '__error__' in log_file.read_text()


def test_rotating_file_handler():
    logger_name = elib.custom_random.random_string()
    log_file = Path(f'./{logger_name}.log')
    assert not log_file.exists()
    logger = elib.custom_logging.get_logger(logger_name, rotate_logs=True, log_to_file=True)
    logger.warning('test')
    assert log_file.exists()
    assert 'test' in log_file.read_text()
    assert len(list(log_file.parent.glob('*'))) == 1
    for handler in logger.handlers:
        if isinstance(handler, logging.handlers.TimedRotatingFileHandler):
            handler.doRollover()
            assert len(list(log_file.parent.glob('*'))) == 2
            break
    else:
        assert logger.handlers is None
        raise LookupError('no rotating file handler found')


# noinspection PyShadowingNames
def test_custom_handler(setup_logging):
    logger, _ = setup_logging

    result = False

    def _callback(*_):
        nonlocal result
        result = True

    handler = elib.custom_logging._custom_logging_handler.CustomLoggingHandler('handler')
    handler.emit = _callback

    handler.register(logger)

    logger.info('test')
    assert result


@pytest.mark.parametrize(
    'level',
    [logging.DEBUG, logging.INFO, logging.ERROR, logging.WARN, logging.CRITICAL]
)
def test_handler_level(level):
    logger = elib.custom_logging.get_logger(elib.custom_random.random_string(4), log_to_file=True)
    elib.custom_logging.set_handler_level(logger.name, 'ch', level)
    elib.custom_logging.set_handler_level(logger.name, 'fh', level)

    for handler in logger.handlers:
        assert handler.level is level

    handler = elib.custom_logging._custom_logging_handler.CustomLoggingHandler('handler')
    handler.register(logger)

    elib.custom_logging.set_handler_level(logger.name, 'handler', level)

    for handler in logger.handlers:
        assert handler.level is level


def test_console_handler_level_as_string():
    logger = elib.custom_logging.get_logger(elib.custom_random.random_string(4), log_to_file=True)

    elib.custom_logging.set_handler_level(logger.name, 'ch', 'critical')
    elib.custom_logging.set_handler_level(logger.name, 'fh', 'critical')
    for handler in logger.handlers:
        assert handler.level is logging.CRITICAL

    elib.custom_logging.set_handler_level(logger.name, 'ch', 'debug')
    elib.custom_logging.set_handler_level(logger.name, 'fh', 'debug')
    for handler in logger.handlers:
        assert handler.level is logging.DEBUG

    elib.custom_logging.set_handler_level(logger.name, 'ch', 'WaRN')
    elib.custom_logging.set_handler_level(logger.name, 'fh', 'warning')
    for handler in logger.handlers:
        assert handler.level is logging.WARN

    elib.custom_logging.set_handler_level(logger.name, 'ch', 'err')
    elib.custom_logging.set_handler_level(logger.name, 'fh', 'error')
    for handler in logger.handlers:
        assert handler.level is logging.ERROR


def test_removal_of_log_file():
    logger_name = elib.custom_random.random_string(4)
    log_file = Path(f'./{logger_name}.log')
    log_file.write_text('dummy content')
    logger = elib.custom_logging.get_logger(logger_name, log_to_file=True)
    assert f'added file logging handler: {log_file.absolute()}' in log_file.read_text()
    assert not 'dummy content' in log_file.read_text()
    logger.debug('some text')
    assert 'some text' in log_file.read_text()


def test_no_log_file():
    logger_name = elib.custom_random.random_string(4)
    log_file = Path(f'./{logger_name}.log')
    logger = elib.custom_logging.get_logger(logger_name, log_to_file=False)
    assert not log_file.exists()
    assert len(logger.handlers) == 1


def test_activate_elib_logging():
    logger = logging.getLogger('some_logger')
    handler = logging.StreamHandler()
    logger.addHandler(handler)
    elib.custom_logging._constants.ROOT_LOGGER = logger
    elib.custom_logging.activate_elib_logging()
    from elib import LOGGER as ELIB_LOGGER
    assert handler in ELIB_LOGGER.handlers


def test_get_elib_logger():
    assert elib.custom_logging.get_elib_logger() is elib.LOGGER


@pytest.mark.parametrize('level', ['debug', 'info', 'warning', 'error', 'critical'])
def test_click_handler(level, capsys):
    logger = elib.custom_logging.get_logger(
        elib.custom_random.random_string(),
        console_level=logging.DEBUG,
        use_click_handler=True
    )
    func = getattr(logger, level)
    func(level)
    out, err = capsys.readouterr()
    if level in ['debug', 'info', 'warning']:
        assert level in out
    else:
        assert level in err



# coding=utf-8
"""
Config property
"""
# pylint: disable=too-few-public-methods

import everett.manager

from .config import BaseConfig


class _ConfigProp:
    """
    Actual descriptor object created during ConfigProp.__call___ below.

    Accessing a ConfigProp from the Config class itself (Config.PROPERTY as opposed to Config().property) gives access
    to the object itself, with "default", "type", "__doc__", "key" and "func".

    """

    def __init__(self, func: callable, default: object, parser: object, namespace):
        """
        Initialize the DESCRIPTOR.

        :param func: callable to overwrite
        :param default: default value if there's nothing in the EverettConfig yet
        :param parser: type of object allowed to be SET
        """
        self.func = func
        self.prop_name = self.func.__name__.upper()
        if namespace:
            self.prop_name = self.prop_name.replace(
                namespace.upper() + '_', '')
        if not isinstance(default, str):
            default = str(default)
        self.default = default
        if parser is bool:
            parser = everett.manager.parse_bool
        self.parser = parser
        self.__doc__ = func.__doc__
        self.namespace = namespace

    def __get__(self, instance, owner=None):
        """
        Retrieves value.

        If instance is None, returns the descriptor object to allow access to "default" and "type"

        If instance is a valid Config object, returns value from it

        :param instance: instance of AbstractMeta
        :param owner: actual Class of the META instance
        :return: value of this DESCRIPTOR (DEFAULT if it isn't defined)
        """

        # raise FileNotFoundError()
        if instance is None:
            # Calling from Class object
            return self

        if not isinstance(instance, BaseConfig):
            raise TypeError(
                '_ConfigProp can only be used with EverettConfig() instances')
        if self.default == 'NO_DEFAULT':
            return getattr(instance, '_config')(
                self.prop_name,
                parser=self.parser,
                namespace=self.namespace
            )

        return getattr(instance, '_config')(
            self.prop_name,
            default=self.default,
            parser=self.parser,
            namespace=self.namespace
        )


class ConfigProp:
    """
    Decorator-class to create properties for META instances.
    """

    def __init__(self, parser: object, default: object = 'NO_DEFAULT', namespace: str = None):
        """
        Initialize properties of the descriptor.

        :param default: default value of the property if it isn't set yet
        :param parser:
        """
        self.default = default
        self.type = parser
        self.namespace = namespace

    def __call__(self, func: callable) -> _ConfigProp:
        """
        Creates a DESCRIPTOR instance for a method of a META instance.

        :param func: function to decorate
        :return: decorated function as a descriptor instance of _MetaProperty
        :rtype: _MetaProperty
        """
        return _ConfigProp(func, self.default, self.type, self.namespace)



# coding=utf-8
from pathlib import Path

import pytest

import elib


def test_get_hash_from_string():
    test_string = 'this is some dummy test string'
    assert elib.hash_.get_hash(test_string) == 'f4c8f848b02d640c75a2385fd315412b'
    assert elib.hash_.get_hash(
        test_string,
        'sha256'
    ) == '556d2c0500f5e1e40925e65d16cc3c46006006e88ffc5051fb3a7f7c1b6af9b1'


def test_get_hash_from_file():
    test_file = Path('./test_file')
    test_file.write_text('this is some dummy test string')
    assert elib.hash_.get_hash(test_file.read_bytes()) == 'f4c8f848b02d640c75a2385fd315412b'
    assert elib.hash_.get_hash(
        test_file.read_bytes(),
        'sha256'
    ) == '556d2c0500f5e1e40925e65d16cc3c46006006e88ffc5051fb3a7f7c1b6af9b1'


def test_get_hash_wrong_method():
    with pytest.raises(RuntimeError):
        elib.hash_.get_hash('test', 'nope')



# coding=utf-8
"""
Download content from the Web
"""
import os
import time

import certifi
import requests
import tqdm
import urllib3  # type: ignore

from elib import LOGGER

from .hash_ import get_hash

REQUESTS_HEADERS = {'User-agent': 'Mozilla/5.0'}


def _get_http_pool():
    return urllib3.PoolManager(cert_reqs=str('CERT_REQUIRED'),
                               ca_certs=certifi.where())


class Downloader:  # pylint: disable=too-many-instance-attributes,too-many-arguments
    """
    Downloads files
    """

    def __init__(
            self,
            url: str,
            filename: str,
            content_length: int = None,
            hexdigest=None,
            download_retries: int = 3,
            block_size: int = 4096 * 4,
            hash_method: str = 'md5',
    ):

        self.url = url
        self.filename = filename
        self.content_length = content_length or None
        self.max_download_retries = download_retries
        self.block_size = block_size
        self.http_pool = _get_http_pool()
        self.hexdigest = hexdigest
        self.file_binary_data = None

        self.hash_method = hash_method

    def _write_to_file(self):

        with open(self.filename, 'wb') as outfile:
            outfile.write(self.file_binary_data)

    def _check_hash(self):

        if self.hexdigest is None:
            LOGGER.debug('no hash to verify')
            return None

        if self.file_binary_data is None:
            LOGGER.debug('cannot verify file hash')
            return False

        LOGGER.debug('checking file hash')
        LOGGER.debug('update hash: %s', self.hexdigest)

        file_hash = get_hash(self.file_binary_data, self.hash_method)

        if file_hash == self.hexdigest:
            LOGGER.debug('file hash verified')
            return True

        LOGGER.debug('cannot verify file hash')
        return False

    @staticmethod
    def _calc_eta(start, now, total, current):  # pragma: no cover

        if total is None:
            return '--:--'

        dif = now - start
        if current == 0 or dif < 0.001:
            return '--:--'

        rate = float(current) / dif
        eta = int((float(total) - float(current)) / rate)
        (eta_mins, eta_secs) = divmod(eta, 60)

        if eta_mins > 99:
            return '--:--'

        return '%02d:%02d' % (eta_mins, eta_secs)

    @staticmethod
    def _calc_progress_percent(received, total):  # pragma: no cover

        if total is None:
            return '-.-%'

        percent = float(received) / total * 100
        percent = '%.1f' % percent

        return percent

    @staticmethod
    def _get_content_length(data):  # pragma: no cover

        content_length = data.headers.get("Content-Length")

        if content_length is not None:
            content_length = int(content_length)

        LOGGER.debug('Got content length of: %s', content_length)

        return content_length

    @staticmethod
    def _best_block_size(time_, chunk: float):  # pragma: no cover

        new_min = max(chunk / 2.0, 1.0)
        new_max = min(max(chunk * 2.0, 1.0), 4194304)  # Do not surpass 4 MB

        if time_ < 0.001:
            return int(new_max)

        rate = chunk / time_

        if rate > new_max:
            return int(new_max)

        if rate < new_min:
            return int(new_min)

        return int(rate)

    def _create_response(self):  # pragma: no cover
        data = None
        LOGGER.debug('Url for request: %s', self.url)

        try:
            data = self.http_pool.urlopen('GET', self.url,
                                          preload_content=False,
                                          retries=self.max_download_retries, )

        except urllib3.exceptions.SSLError:
            LOGGER.debug('SSL cert not verified')

        except urllib3.exceptions.MaxRetryError:
            LOGGER.debug('MaxRetryError')

        except Exception as exc:  # pylint: disable=broad-except
            LOGGER.debug(str(exc), exc_info=True)

        if data is not None:
            LOGGER.debug('resource URL: %s', self.url)
        else:
            LOGGER.debug('could not create resource URL.')
        return data

    def download_to_memory(self):
        """
        Download bytes to memory
        """

        data = self._create_response()

        if data is None:
            return None

        self.content_length = self._get_content_length(data)

        if self.content_length is None:  # pragma: no cover
            LOGGER.debug('content-Length not in headers')
            LOGGER.debug('callbacks will not show time left '
                         'or percent downloaded.')

        received_data = 0

        start_download = time.time()
        block = data.read(1)
        received_data += len(block)
        self.file_binary_data = block
        percent = self._calc_progress_percent(0, self.content_length)

        # with click.progressbar(length=self.content_length, label=f'Downloading {self.url}') as progress:
        with tqdm.tqdm(total=self.content_length, unit_scale=True, unit='B',
                       desc=f'Downloading {self.url}') as progress:

            current = 0

            def _progress_hook(data_):
                nonlocal current
                progress.update(data_['downloaded'] - current)
                current = data_['downloaded']
                # progress.label = data['time']

            while 1:

                start_block = time.time()

                block = data.read(self.block_size)

                end_block = time.time()

                if not block:
                    break

                self.block_size = self._best_block_size(end_block - start_block, len(block))

                self.file_binary_data += block

                received_data += len(block)

                percent = self._calc_progress_percent(received_data,
                                                      self.content_length)

                time_left = self._calc_eta(start_download, time.time(),
                                           self.content_length,
                                           received_data)

                status = {'total': self.content_length,
                          'downloaded': received_data,
                          'status': 'downloading',
                          'percent_complete': percent,
                          'time': time_left}

                _progress_hook(status)

            status = {'total': self.content_length,
                      'downloaded': received_data,
                      'status': 'finished',
                      'percent_complete': percent,
                      'time': '00:00'}

            _progress_hook(status)
        LOGGER.debug('Download Complete')

    def download(self) -> bool:
        """
        Download content to file

        Returns: success of the operation

        """
        LOGGER.debug('downloading to memory')
        self.download_to_memory()

        check = self._check_hash()

        if check is True or check is None:
            LOGGER.debug('writing to file')
            self._write_to_file()
            return True

        else:
            del self.file_binary_data
            if os.path.exists(self.filename):
                try:
                    os.remove(self.filename)
                except OSError:  # pragma: no cover
                    pass
            return False


# TODO: move to elib
def download(
        url: str,
        outfile: str,
        hexdigest=None,
) -> bool:
    """
    Download file

    Args:
        url: source
        outfile: local file to save the content to
        hexdigest: optional hexdigest to check the download

    Returns: success of the operation

    """
    LOGGER.info(f'downloading: {locals()}')
    req = requests.head(url, headers=REQUESTS_HEADERS, timeout=5)
    if not req.ok:
        LOGGER.error(f'Download failed: {req.reason}')
        return False

    LOGGER.debug('Processing download request')

    return Downloader(
        url=url,
        filename=outfile,
        hexdigest=hexdigest,
    ).download()



# coding=utf-8
"""
Tests path package
"""

from pathlib import Path

import pytest

import elib.path


def test_path():
    path_str = './test'
    path_test = Path('./test')
    assert elib.path.ensure_path(path_str, must_exist=False) == elib.path.ensure_path(path_test, must_exist=False)
    with pytest.raises(FileNotFoundError):
        elib.path.ensure_path(path_str)
    with pytest.raises(FileNotFoundError):
        elib.path.ensure_path(path_test)


def test_dir():
    test = Path('./test')
    assert isinstance(elib.path.ensure_dir(test, must_exist=False), Path)
    test.touch()
    with pytest.raises(TypeError):
        elib.path.ensure_dir(test)
    test.unlink()
    test.mkdir()
    assert isinstance(elib.path.ensure_dir(test), Path)


def test_file():
    test = Path('./test')
    assert isinstance(elib.path.ensure_file(test, must_exist=False), Path)
    test.mkdir()
    with pytest.raises(TypeError):
        elib.path.ensure_file(test)
    test.rmdir()
    test.touch()
    assert isinstance(elib.path.ensure_file(test), Path)


@pytest.mark.parametrize('first', ['some', Path('some')])
@pytest.mark.parametrize('second', ['path', Path('path')])
@pytest.mark.parametrize('third', ['here', Path('here')])
def test_args(first, second, third):
    test = elib.path.ensure_path(first, second, third, must_exist=False)
    assert isinstance(test, Path)
    assert test == Path('some/path/here').absolute()


def test_dir_create():
    path = 'some.dir'
    test = Path(path).absolute()
    assert not test.exists()
    result = elib.path.ensure_dir(path, must_exist=False, create=True)
    assert test.exists()
    assert test == result



# coding=utf-8
"""
Config test
"""
from pathlib import Path

import everett
import pytest

from elib.config import BaseConfig
# noinspection PyProtectedMember
from elib.config.property import ConfigProp, _ConfigProp


class WrongBaseClass:
    """
    Does not inherit BaseConfig
    """
    @ConfigProp(str, '')
    def string(self):
        """Dummy prop"""
        pass


class DummyConfig(BaseConfig):
    """Dummy test class for Config"""
    @ConfigProp(bool, False)
    def debug(self):
        """Dummy prop"""
        pass

    @ConfigProp(str, '')
    def string(self):
        """Dummy prop"""
        pass

    @ConfigProp(int, 0)
    def integer(self):
        """Dummy prop"""
        pass

    @ConfigProp(list, None)
    def some_list(self):
        """Dummy prop"""
        pass

    @ConfigProp(str, namespace='namespace')
    def namespace_key(self):
        """Dummy prop"""
        pass

    @ConfigProp(str)
    def no_default(self):
        """Dummy prop"""
        pass


def test_create_config():
    cfg = DummyConfig('test')
    assert cfg.debug is False
    assert not cfg.string
    assert cfg.integer is 0


def test_ini_config_file():
    with open('test.ini', 'w') as stream:
        stream.write('''
[main]
debug=true
string=some string
integer=1
''')
    cfg = DummyConfig('test')
    assert cfg.debug is True
    assert cfg.string == 'some string'
    assert cfg.integer is 1


@pytest.mark.parametrize('ext', ['yaml', 'yml'])
def test_yaml_config_file(ext):
    with open(f'test.{ext}', 'w') as stream:
        stream.write('''
debug: "true"
string: some string
integer: 1
''')
    cfg = DummyConfig('test')
    assert cfg.debug is True
    assert cfg.string == 'some string'
    assert cfg.integer is 1


@pytest.mark.parametrize('key', ['debug', 'DEBUG'])
def test_default_dict(key):
    cfg = DummyConfig(
        'test', {key: 'true', 'string': 'some string', 'integer': 1})
    assert cfg.debug is True
    assert cfg.string == 'some string'
    assert cfg.integer is 1


@pytest.mark.parametrize('ext', ['yaml', 'yml'])
def test_yaml_list_of_str(ext):
    with open(f'test.{ext}', 'w') as stream:
        stream.write('''
debug: "true"
string: some string
integer: 1
some_list:
  - caribou
  - gopher
  - moose
''')
    cfg = DummyConfig('test')
    assert cfg.some_list == ['caribou', 'gopher', 'moose']


@pytest.mark.parametrize('ext', ['yaml', 'yml'])
def test_yaml_list_of_int(ext):
    with open(f'test.{ext}', 'w') as stream:
        stream.write('''
debug: "true"
string: some string
integer: 1
some_list:
  - 0
  - -4
  - 3
''')
    cfg = DummyConfig('test')
    assert cfg.some_list == [0, -4, 3]


@pytest.mark.parametrize('ext', ['yaml', 'yml'])
def test_yaml_namespace(ext):
    with open(f'test.{ext}', 'w') as stream:
        stream.write('''
debug: "true"
string: some string
integer: 1
some_list:
  - caribou
  - gopher
  - moose
namespace:
  key: 'caribou'
''')
    cfg = DummyConfig('test')
    assert cfg.namespace_key == 'caribou'


def test_no_default():
    cfg = DummyConfig('test')
    with pytest.raises(everett.ConfigurationMissingError):
        assert cfg.no_default == 'value'


def test_wrong_base_class():
    test = WrongBaseClass()
    with pytest.raises(TypeError):
        print(test.string)


def test_calling_from_instance():
    assert isinstance(DummyConfig.integer, _ConfigProp)


def test_empty_config_file():
    Path('./test.yml').touch()
    cfg = DummyConfig('test')
    assert cfg.debug is False



# coding=utf-8
"""
Convenience functions to manage logging
"""
from elib.custom_logging._custom_logging_handler import CustomLoggingHandler
from ._custom_logging import activate_elib_logging, get_logger, set_handler_level, get_elib_logger



# coding=utf-8
"""
Manages Config
"""
# pylint: disable=too-few-public-methods

import os

import everett
import everett.manager

from .yaml_config import YAMLConfig


class BaseConfig:
    """
    Singleton configuration class for EDLM.
    """

    def __init__(self, package_name: str, default_dict: dict = None):
        if default_dict is None:
            default_dict = {}
        self._config = everett.manager.ConfigManager(
            [
                everett.manager.ConfigEnvFileEnv('.env'),
                everett.manager.ConfigOSEnv(),
                YAMLConfig(
                    [
                        os.environ.get(f'{package_name.upper()}_YAML'),
                        os.path.join(os.path.expanduser(
                            '~'), f'{package_name}.yml'),
                        os.path.join(os.path.expanduser('~'),
                                     f'{package_name}.yaml'),
                        f'./{package_name}.yml',
                        f'./{package_name}.yaml',
                    ]
                ),
                everett.manager.ConfigIniEnv(
                    [
                        os.environ.get(f'{package_name.upper()}_INI'),
                        os.path.join(os.path.expanduser(
                            '~'), f'{package_name}.ini'),
                        f'./{package_name}.ini',
                    ]
                ),
                everett.manager.ConfigDictEnv(default_dict),
            ]
        )



# coding=utf-8

import sys
from pathlib import Path

import elib._run
from elib.settings import ELIBSettings


def test_find_executable():
    python = elib._run.find_executable('python')
    assert elib._run.find_executable('python.exe') == python
    assert elib._run.find_executable('python', f'{sys.prefix}/Scripts') == python
    assert elib._run.find_executable('__sure__not__') is None


def test_context():
    assert elib._run.find_executable('__sure__not__') is None
    ELIBSettings.known_executables['__sure__not__.exe'] = 'ok'
    assert elib._run.find_executable('__sure__not__') == 'ok'


def test_paths():
    assert elib._run.find_executable('python')
    assert elib._run.find_executable('python', '.')
    ELIBSettings.known_executables = {}
    assert elib._run.find_executable('python', '.') is None


def test_immediate():
    Path('./test.exe').touch()
    assert elib._run.find_executable('test')



# coding=utf-8
"""
Outputs log message via click
"""

import logging

import click


class ClickHandler(logging.Handler):
    """
    Outputs log message via click
    """

    def __init__(self):
        logging.Handler.__init__(self)

    @staticmethod
    def _debug(msg):
        click.echo(click.style(msg, dim=True))

    @staticmethod
    def _info(msg):
        click.echo(click.style(msg, fg='cyan'))

    @staticmethod
    def _warning(msg):
        click.echo(click.style(msg, bg='yellow'))

    @staticmethod
    def _error(msg):
        click.echo(click.style(msg, bg='red'), err=True)

    @staticmethod
    def _critical(msg):
        click.echo(click.style(msg, bg='red'), err=True)

    def emit(self, record: logging.LogRecord):
        """
        Do whatever it takes to actually log the specified logging record.
        """
        _func_map = {
            logging.DEBUG: self._debug,
            logging.INFO: self._info,
            logging.WARNING: self._warning,
            logging.ERROR: self._error,
            logging.CRITICAL: self._critical,
        }
        msg = self.format(record)
        _func_map[record.levelno](msg)



# coding=utf-8
"""
Utilities to obtain random values
"""

import os
import random
import string


def random_string(size=6, chars=string.ascii_uppercase + string.digits) -> str:
    """
    Creates a random string

    :param size: len of str
    :param chars: character pool
    :return: random string
    """
    return ''.join(random.choice(chars) for _ in range(size))


def random_bytes(length=1024):
    """
    Creates a random bytes chain

    Args:
        length: size of the chain

    Returns: bytes

    """
    return os.urandom(length)



# coding=utf-8
"""
Pass
"""

import string

import pytest
from hypothesis import given
from hypothesis import strategies as st

import elib.console
from elib.settings import ELIBSettings


@given(text=st.text())
def test_sanitize(text):
    result = elib.console._sanitize(text)
    assert isinstance(result, str)
    result.encode('ascii')


@given(text=st.text(alphabet=string.printable))
def test_info(text, capsys):
    elib.console.info(text)
    out, err = capsys.readouterr()
    assert out == 'EPAB: {}\n'.format(text)
    assert err == ''


@given(text=st.text(alphabet=string.printable))
def test_cmd_end(text, capsys):
    elib.console.cmd_end(text)
    out, err = capsys.readouterr()
    assert out == '{}\n'.format(text)
    assert err == ''


# @given(text=st.text(alphabet=string.printable))
# def test_cmd_start(text, capsys):
#     elib.console.cmd_start(text)
#     out, err = capsys.readouterr()
#     assert out == 'EPAB: {}'.format(text)
#     assert err == ''


# @given(text=st.text(alphabet=string.printable))
# def test_std_err(text, capsys):
#     elib.console.std_err(text)
#     out, err = capsys.readouterr()
#     assert out == ''
#     assert err == 'EPAB: {}\n'.format(text)


# @given(text=st.text(alphabet=string.printable))
# def test_std_out(text, capsys):
#     elib.console.std_out(text)
#     out, err = capsys.readouterr()
#     assert out == text
#     assert err == ''


# @given(text=st.text(alphabet=string.printable))
# def test_error(text, capsys):
#     elib.console.error(text)
#     out, err = capsys.readouterr()
#     assert out == ''
#     assert err == 'EPAB: {}\n'.format(text)


@pytest.mark.parametrize(
    'func,out,err',
    [
        (elib.console.info, 'EPAB: {}\n', ''),
        (elib.console.error, '', 'EPAB: {}\n'),
        (elib.console.cmd_start, 'EPAB: {}', ''),
        (elib.console.cmd_end, '{}\n', ''),
        (elib.console.std_err, '', '{}\n'),
        (elib.console.std_out, '{}', ''),
    ],
    ids=['info', 'error', 'cmd_start', 'cmd_end', 'std_err', 'std_out']
)
@given(text=st.text(alphabet=string.printable))
def test_quiet(func, out, err, text, capsys):
    ELIBSettings.quiet = False
    func(text)
    _out, _err = capsys.readouterr()
    assert _out == out.format(text)
    assert _err == err.format(text)
    ELIBSettings.quiet = True
    func(text)
    out, err = capsys.readouterr()
    assert out == ''
    assert err == ''



# coding=utf-8
"""
ELIB settings
"""


class ELIBSettings:
    """
    ELIB settings
    """

    mute: bool = False
    quiet: bool = False
    verbose: bool = False
    known_executables: dict = {}

    color_info = 'green'
    color_error = 'red'
    color_cmd = 'magenta'
    color_stdout = 'cyan'
    color_stderr = 'red'



# coding=utf-8
from pathlib import Path

import requests
from mockito import mock, verify, when

from elib import downloader

URL = r'http://www.ovh.net/files/1Mio.dat'


def test_download():
    assert downloader.download(url=URL, outfile='./test', hexdigest='6cb91af4ed4c60c11613b75cd1fc6116')
    assert Path('./test').exists()


def test_download_wrong_digest():
    assert not downloader.download(url=URL, outfile='./test', hexdigest='nope')
    assert not Path('./test').exists()


def test_download_no_digest():
    assert downloader.download(url=URL, outfile='./test')
    assert Path('./test').exists()


def test_download_no_data():
    when(downloader.Downloader).download_to_memory(...)
    assert not downloader.download(url=URL, outfile='./test', hexdigest='6cb91af4ed4c60c11613b75cd1fc6116')
    verify(downloader.Downloader).download_to_memory(...)
    assert not Path('./test').exists()


def test_download_to_memory_no_data():
    when(downloader.Downloader)._create_response().thenReturn(None)
    assert not downloader.download(url=URL, outfile='./test', hexdigest='6cb91af4ed4c60c11613b75cd1fc6116')
    verify(downloader.Downloader)._create_response()
    assert not Path('./test').exists()


def test_download_delete_failed():
    Path('./test').touch()
    assert Path('./test').exists()
    when(downloader.Downloader)._create_response().thenReturn(None)
    assert not downloader.download(url=URL, outfile='./test', hexdigest='6cb91af4ed4c60c11613b75cd1fc6116')
    verify(downloader.Downloader)._create_response()
    assert not Path('./test').exists()


def test_download_request_failed():
    req = mock()
    req.ok = False
    req.reason = 'nope'
    when(requests).head(...).thenReturn(req)
    assert not downloader.download(url=URL, outfile='./test', hexdigest='6cb91af4ed4c60c11613b75cd1fc6116')
    verify(requests).head(...)
    assert not Path('./test').exists()



# coding=utf-8
"""
This is a temporary test to cover the import of modules that have 0% coverage so far

Yes, this is shameful...
"""

import glob

import pytest


# noinspection PyUnresolvedReferences,PyProtectedMember
@pytest.mark.nocleandir
@pytest.mark.parametrize('module_', glob.glob('./elib/**/*.py', recursive=True))
def test_imports(module_):
    module_ = module_[2:-3].replace('\\', '.')
    __import__(module_)


# noinspection PyUnresolvedReferences,PyProtectedMember
@pytest.mark.nocleandir
@pytest.mark.parametrize('module_', list(glob.glob('./test/**/*.py', recursive=True)))
def test_imports_tests(module_):
    module_ = module_[2:-3].replace('\\', '.')
    __import__(module_)



# coding=utf-8
"""
Pytest config
"""
import os
import sys

import pytest
from mockito import unstub


# noinspection PyUnusedLocal
def pytest_configure(config):
    """Setup"""
    sys._called_from_test = True


# noinspection PyUnusedLocal,SpellCheckingInspection
def pytest_unconfigure(config):
    """Tear down"""
    # noinspection PyUnresolvedReferences,PyProtectedMember
    del sys._called_from_test


@pytest.fixture(autouse=True)
def cleandir(request, tmpdir):
    """Provides a clean working dir"""
    if 'nocleandir' in request.keywords:
        yield
    else:
        current_dir = os.getcwd()
        os.chdir(str(tmpdir))
        yield os.getcwd()
        os.chdir(current_dir)


@pytest.fixture(autouse=True)
def _unstub():
    unstub()
    yield
    unstub()


def pytest_addoption(parser):
    """Add option for long tests"""
    parser.addoption("--long", action="store_true",
                     help="run long tests")


def pytest_runtest_setup(item):
    """Skip long tests"""
    long_marker = item.get_marker("long")
    if long_marker is not None and not item.config.getoption('long'):
        pytest.skip('skipping long tests')



# coding=utf-8
"""
Etcher's stupid library
"""

import os

from pip.req import parse_requirements
from setuptools import find_packages, setup

requirements = [str(r.req) for r in
                parse_requirements('requirements.txt', session=False)]
test_requirements = [str(r.req) for r in
                     parse_requirements('requirements-dev.txt', session=False)]

CLASSIFIERS = filter(None, map(str.strip,
                               """
Development Status :: 3 - Alpha
Environment :: Win32 (MS Windows)
Intended Audience :: Developers
Natural Language :: English
Operating System :: Microsoft :: Windows :: Windows 7
Operating System :: Microsoft :: Windows :: Windows 8
Operating System :: Microsoft :: Windows :: Windows 8.1
Operating System :: Microsoft :: Windows :: Windows 10
License :: OSI Approved :: MIT License
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Topic :: Utilities
""".splitlines()))


def read_local_files(*file_paths: str) -> str:
    """
    Reads one or more text files and returns them joined together.

    A title is automatically created based on the file name.

    :param file_paths: list of files to aggregate

    """

    def _read_single_file(file_path):
        with open(file_path) as f:
            filename = os.path.splitext(file_path)[0]
            title = f'{filename}\n{"=" * len(filename)}'
            return '\n\n'.join((title, f.read()))

    return '\n' + '\n\n'.join(map(_read_single_file, file_paths))


setup(
    name='elib',
    zip_safe=False,
    install_requires=requirements,
    tests_require=test_requirements,
    package_dir={'elib': 'elib'},
    package_data={},
    test_suite='pytest',
    packages=find_packages(),
    python_requires='>=3.6',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    license='MIT',
    classifiers=CLASSIFIERS,
)



# coding=utf-8
"""
Manages runners
"""

import os
import sys
import typing
from pathlib import Path

import click
import delegator

from elib.console import cmd_end, cmd_start, error, info, std_err, std_out
from elib.settings import ELIBSettings

KNOWN_EXECUTABLES = {}


def _append_exe(executable):
    if not executable.endswith('.exe'):
        return f'{executable}.exe'

    return executable


def _set_paths(*paths: str):
    if not paths:
        path = os.environ['PATH']
        paths = [Path(sys.exec_prefix, 'Scripts').absolute()] + path.split(os.pathsep)

    return paths


def _search_paths(paths, executable):
    for path_ in paths:
        executable_path = Path(path_, executable).absolute()
        if executable_path.is_file():
            break
    else:
        cmd_end(f' -> not found')
        return None

    return executable_path


def find_executable(executable: str, *paths: str) -> typing.Optional[Path]:  # noqa: C901
    # noinspection SpellCheckingInspection
    """
    https://gist.github.com/4368898

    Public domain code by anatoly techtonik <techtonik@gmail.com>

    Programmatic equivalent to Linux `which` and Windows `where`

    Find if ´executable´ can be run. Looks for it in 'path'
    (string that lists directories separated by 'os.pathsep';
    defaults to os.environ['PATH']). Checks for all executable
    extensions. Returns full path or None if no command is found.

    Args:
        executable: executable name to look for
        paths: root paths to examine (defaults to system PATH)

    Returns: executable path as string or None

    """

    executable = _append_exe(executable)

    if executable in ELIBSettings.known_executables:  # type: ignore
        return ELIBSettings.known_executables[executable]  # type: ignore

    cmd_start(f'Looking for executable: {executable}')

    paths = _set_paths(*paths)

    executable_path = Path(executable)
    if not executable_path.is_file():
        executable_path = _search_paths(paths, executable)

    if executable_path is None:
        cmd_end(f' -> not found')
        return None

    ELIBSettings.known_executables[executable] = executable_path
    cmd_end(f' -> {click.format_filename(str(executable_path))}')
    return executable_path


def filter_line(
        line: str,
        filters: typing.Optional[typing.Iterable[str]]
) -> typing.Optional[str]:
    """
    Filters out line that contain substring

    Args:
        line: line to filter
        filters: filter strings to apply

    Returns: line if not filter are found in line, else None

    """
    if filters is not None:
        for filter_ in filters:
            if filter_ in line:
                return None
    return line


def _parse_output(process, filters):
    result = []
    for line in process.out.splitlines():
        if filter_line(line, filters):
            result.append(line)
    for line in process.err.splitlines():  # pragma: no cover
        if filter_line(line, filters):
            result.append(line)

    return '\n'.join(result)


def _process_run_error(
        mute,
        result,
        failure_ok,
        process,
        exe_short
):
    if mute:
        cmd_end('')
    error(f'command failed: {exe_short} -> {process.return_code}')
    if result:
        std_err(f'{exe_short} error:\n{result}')
        if not result.endswith('\n'):  # pragma: no cover
            print()
    if not failure_ok:
        exit(process.return_code)


def _process_run_success(
        mute,
        result,
        process,
        exe_short
):
    if mute:
        cmd_end(f' -> {process.return_code}')
    else:
        std_out(result)
        if not result.endswith('\n'):  # pragma: no cover
            print()
        info(f'{exe_short} -> {process.return_code}')


def _process_run_result(process, mute, exe_short, failure_ok, result) -> typing.Tuple[str, int]:
    if process.return_code:
        _process_run_error(
            mute=mute,
            result=result,
            failure_ok=failure_ok,
            process=process,
            exe_short=exe_short,
        )
    else:
        _process_run_success(
            mute=mute,
            result=result,
            process=process,
            exe_short=exe_short,
        )

    return result, process.return_code


def _prepare_run_setup_filters(filters):
    if filters and isinstance(filters, str):
        filters = [filters]

    return filters


def _prepare_run_find_exe(cmd, *paths):
    exe = find_executable(cmd.split(' ')[0], *paths)
    if not exe:
        exit(-1)
    return exe


def _prepare_run_advertise(mute, cmd):
    mute = mute and not ELIBSettings.verbose

    if mute:
        cmd_start(f'RUNNING: {cmd}')
    else:
        info(f'RUNNING: {cmd}')


def run(
        cmd: str,
        *paths: str,
        cwd: str = '.',
        mute: bool = False,
        filters: typing.Union[None, typing.Iterable[str]] = None,
        failure_ok: bool = False,
) -> typing.Tuple[str, int]:
    """
    Executes a command and returns the result

    Args:
        cmd: command to execute
        paths: paths to search executable in
        cwd: working directory (defaults to ".")
        mute: if true, output will not be printed
        filters: gives a list of partial strings to filter out from the output (stdout or stderr)
        failure_ok: if False (default), a return code different than 0 will exit the application

    Returns: command output
    """

    filters = _prepare_run_setup_filters(filters)

    exe = _prepare_run_find_exe(cmd, *paths)
    exe_short = exe.name

    cmd = ' '.join([f'"{exe.absolute()}"'] + cmd.split(' ')[1:])

    _prepare_run_advertise(mute, cmd)

    process = delegator.run(cmd, block=True, cwd=cwd, binary=False)
    result = _parse_output(process, filters)

    return _process_run_result(process, mute, exe_short, failure_ok, result)



# coding=utf-8

import elib


def test_pretty():
    obj = {
        'test': 'value',
        'nested': {
            'test': 'value'
        },
        'a': 't',
        'b': 't',
        'c': 't',
        'd': 't',
        'e': 't',
        'f': 't',
        'g': 't',
        'h': 't',
        'i': 't',
        'j': 't',
        'k': 't',
        'l': 't',
    }
    out = elib.pretty_format(obj)
    print(out)
    assert out == """    {'a': 't',
     'b': 't',
     'c': 't',
     'd': 't',
     'e': 't',
     'f': 't',
     'g': 't',
     'h': 't',
     'i': 't',
     'j': 't',
     'k': 't',
     'l': 't',
     'nested': {'test': 'value'},
     'test': 'value'}"""


def test_pretty_no_indent():
    obj = {
        'test': 'value'
    }
    out = elib.pretty_format(obj, indent=0)
    assert out == "{'test': 'value'}"



# coding=utf-8
"""
Creates a snippet at glot.io for text sharing
"""

import json
import typing
from collections import namedtuple

import requests

BASE_URL = 'https://snippets.glot.io/snippets'

PasteContent = namedtuple('PasteContent', 'filename content')


class PasteError(requests.HTTPError):
    """Raised on paste error"""
    pass


def create_new_paste(title, files: typing.List[PasteContent], public: bool = False, language='lua'):
    """
    Creates a paste on Glot

    :param title: name of the paste
    :param files: list of PasteContent
    :param public: whether or not paste is public
    :param language: syntax highlighting
    :return: URL of the newly created paste
    """
    headers = {
        'Content-type': 'application/json',
    }
    data = {
        'language': language,
        'title': title,
        'public': public,
        'files': [
            {
                'name': file.filename,
                'content': file.content
            } for file in files
        ],
    }
    req = requests.post(BASE_URL, json=data, headers=headers)
    if req.ok:
        resp = json.loads(req.text)
        url = resp['url'].replace('https://snippets.glot.io', 'https://glot.io')
        return url
    else:
        raise PasteError(f'failed to post content; reason: {req.reason}')



# coding=utf-8
"""
Manages TTS utilities
"""
import typing
from pathlib import Path

import gtts

import elib.custom_logging
import elib.path
from elib import LOGGER


def text_to_speech(text: str, file_path: typing.Union[str, Path], overwrite: bool = False) -> Path:
    """
    Creates MP3 file from text

    Args:
        text: text to encode
        file_path: path to MP3 file
        overwrite: whether or not to overwrite existing file

    Returns: path to saved MP3

    """
    LOGGER.debug(f'{text}\n->{file_path}')
    file_path = elib.path.ensure_path(file_path, must_exist=False)
    if file_path.exists() and not overwrite:
        LOGGER.error(f'"{file_path}" already exists')
        raise FileExistsError(file_path)
    LOGGER.debug('encoding text')
    tts = gtts.gTTS(text=text, lang='en', slow=False)
    LOGGER.debug('saving MP3 file')
    tts.save(str(file_path))
    return file_path
